/* given a pole zero response, computes the complex values as chosen by user 
 * This is essentially an educational tool to try and visualize an instrument response.
 * in order to compare with the full response (digitizer + instrument + decimation) use the evalresp
 *   library from IRIS
 */
#include "station_analysis_tools.h"

int main (int argc, char *argv[]) {
	/* local variables */
	unsigned int re_flag = 0, im_flag = 0, am_flag = 0, ph_flag = 0, l=0;
	unsigned int n_freq=1000, i;
	unsigned int unit_flag = 0;
	double min_freq=0.001, max_freq=0.5, freq_inc, freq;
	double *arr_out;
	char paramfile_name[100], pole_zero_file_name[100];
	char out_name[110];
	char buff[100];
	FILE *paramfile, *out;
	POLE_ZERO pz;
	SPECTRUM response_spectrum;
	COMPLEX_RP temp_crp;

	/* check usage */
	if (argc != 2) {
		printf("Usage: %s paramfile\n",argv[0]);
		printf("paramfile format:\n");
		printf("input_pole_zero_file\n");
		printf("min_frequency max_frequency number_of_frequencies\n");
		printf("unit_flag\n");
		printf("re_flag im_flag am_flag ph_flag\n\n\n");
		printf("The unit_flag can be 0 (take the given 0's), 1 (subtract one - velocity response), or 2 (subtract 2 - acceleration)\n");
		printf("the various flags at the end determine if the program will output files with the real, imaginary, amplitude, and phase information respectively.\nEnter 0 for no or 1 for yes. Default is 0 for all\n");
		printf("Default output files append _re, _im, _am, or _ph to the input pole zero name\n");
		exit(1);
	}

	/* -h help */
	if (!(strcmp(argv[1],"-h"))) {
		printf("Usage: %s paramfile\n",argv[0]);
        printf("paramfile format:\n");
        printf("input_pole_zero_file\n");
        printf("min_frequency max_frequency number_of_frequencies\n");
		printf("unit_flag\n");
        printf("re_flag im_flag am_flag ph_flag\n\n\n");
		printf("The unit_flag can be 0 (take the given 0's), 1 (subtract one - velocity response), or 2 (subtract 2 - acceleration)\n");
        printf("the various flags at the end determine if the program will output files with the real, imaginary, amplitude, and phase information respectively.\nEnter 0 for no or 1 for yes. Default is 0 for all\n"); 
        printf("Default output files append _re, _im, _am, or _ph to the input pole zero name\n");
		exit(1);
	}       

	/* get the paramfile name from the argument */
	sscanf(argv[1],"%s",paramfile_name);

	/* try to open the paramfile */
	if ( (paramfile = fopen(paramfile_name,"r")) == NULL) {
		fprintf(stderr,"Error, file %s not found\n",paramfile_name);
		exit(1);
	}

	/* read the parameter file */
	l=0;
	while (fgets(buff,100,paramfile)) {
		if (l == 0) {
			sscanf(buff,"%s",pole_zero_file_name);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%lf %lf %d\n",&min_freq, &max_freq, &n_freq);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%d",&unit_flag);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%d %d %d %d\n",&re_flag, &im_flag, &am_flag, &ph_flag);
			l++;
		}
	}

	/* close the parameter file */
	fclose(paramfile);

	/* reopen a paramfile to write the configuration information to */
	paramfile = fopen("compute_pole_zero_response.cfg","w");
	fprintf(paramfile,"%s\n%lf %lf %d\n%d\n%d %d %d %d\n",pole_zero_file_name,min_freq,max_freq,n_freq,unit_flag,re_flag,im_flag,am_flag,ph_flag);
	fclose(paramfile);
	fprintf(stdout,"Wrote configuration file compute_pole_zero_response.cfg in current directory\n");

	/* read the pole zero file */
	if (read_sac_pole_zero_file(pole_zero_file_name, &pz) == 0 ) {
		fprintf(stderr,"Error, file %s not found\n",pole_zero_file_name);
		exit(1);
	}
	
	/* subtract the appropriate number of zeros */
	pz.n_zeroes = pz.n_zeroes - unit_flag;
	/* check this didn't go to 0 or negative */
	if (pz.n_zeroes < 1) {
		printf("Error, unit_flag option resulted in %d zeros. Value must be 1 or greater!\n",pz.n_zeroes);
		exit(1);
	}


	/* determine the step in frequency */
	if (n_freq > 1) {
		freq_inc = (max_freq - min_freq) / ( n_freq - 1);
	} else {
		fprintf(stderr,"Error, must have more than 1 frequency!\n");
		exit(1);
	}

	/* things have been read, so allocate memory */
	if ( (arr_out = malloc(sizeof(*arr_out) * n_freq)) == NULL) {
		printf("Malloc error with array_out\n");
		exit(1);
	}
	
	/* calls malloc for real and imaginary components. Uses (n_freq-1) * 2 because subroutine assumes the second argument is the number of points in the corresponding time series and this has no time series */
	initialize_spectrum(&response_spectrum,(n_freq-1)*2);
	
	/* loop over each frequency and generate the response putting into a spectrum to be used for amplitude and phase information */
	for (i=0; i<n_freq; i++) {
		freq = i * freq_inc + min_freq;
		temp_crp = generate_response(&pz,freq);
		response_spectrum.re[i] = temp_crp.real;
		response_spectrum.im[i] = temp_crp.imag;
	}
	
	/* if the real component is wanted...*/
	if (re_flag == 1) {
		sprintf(out_name,"%s_re",pole_zero_file_name);
		out = fopen(out_name,"w");
		for (i=0; i<n_freq; i++) {
			freq = i * freq_inc + min_freq;
			fprintf(out,"%lf %lf\n",freq,response_spectrum.re[i]);
		}
		fclose(out);
	}

	/* if the imaginary component is wanted...*/
	if (im_flag == 1) {
		sprintf(out_name,"%s_im",pole_zero_file_name);
		out = fopen(out_name,"w");
		for (i=0; i<n_freq; i++) {
			freq = i * freq_inc + min_freq;
			fprintf(out,"%lf %lf\n",freq,response_spectrum.im[i]);
		}
		fclose(out);
	}

	/* if the amplitude is wanted...*/
	if (am_flag == 1) {
		sprintf(out_name,"%s_am",pole_zero_file_name);
		out = fopen(out_name,"w");
		get_amplitude_spectrum(&response_spectrum,arr_out);
		for (i=0; i<n_freq; i++) {
			freq = i * freq_inc + min_freq;
			fprintf(out,"%lf %lf\n",freq,arr_out[i]);
		}
		fclose(out);
	}

	/* if the phase is wanted...*/
	if (ph_flag == 1) {
		sprintf(out_name,"%s_ph",pole_zero_file_name);
		out = fopen(out_name,"w");
		get_phase_spectrum(&response_spectrum,arr_out);
		for (i=0; i<n_freq; i++) {
			freq = i * freq_inc + min_freq;
			fprintf(out,"%lf %lf\n",freq,arr_out[i]);
		}
		fclose(out);
	}

	/* free memory like a good citizen */
	free(arr_out);
	free(response_spectrum.re);
	free(response_spectrum.im);


	return 0;
}







