/* get_mean_general_pdf.c
 * actually extracts the highest likelihood at each x point
 * more robust estimates can be made by looking at various percentiles
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

int main (int argc, char *argv[]) {
	/* local variables */
	int i, j, l, n_x, y_node, x=0, y=0;
	int max_index=0;
	double max_pdf;
	double **hit_matrix;
	double ymin, ymax, yinc;	
	double *x_array, *out_array, *y_array;
	char paramfile_name[100], input_pdf_name[100];
	char output_file_name[100], buff[100];
	FILE *paramfile, *output_file, *pdf_file;

	/* help usage */
	if (argc == 2) {
		if (!(strcmp(argv[1],"-h"))) {
			fprintf(stderr,"Usage: %s parameter_file\n", argv[0]);
			fprintf(stderr,"parameter_file format: \n");
 	        fprintf(stderr,"input_pdf\n");
            fprintf(stderr,"n_x\n");
            fprintf(stderr,"ymin ymax yinc\n");
            fprintf(stderr,"output_file\n");
            exit(1);
		}
	}

	/* check usage */
	if (argc != 2) {
		fprintf(stderr,"Usage: %s parameter_file\n", argv[0]);
		fprintf(stderr,"parameter_file format: \n");
		fprintf(stderr,"input_pdf\n");
		fprintf(stderr,"n_x\n");
		fprintf(stderr,"ymin ymax yinc\n");
		fprintf(stderr,"output_file\n");
		exit(1);
	}

	/* open parameter file */
	sscanf(argv[1],"%s",paramfile_name);
	if (! (paramfile = fopen(paramfile_name,"r"))) {
		fprintf(stderr,"Error, file %s not found\n",paramfile_name);
		exit(1);
	}

	/* read parameter file */
	l=0;
	while (fgets(buff,100,paramfile)) {
		if (l == 0) {
			sscanf(buff,"%s",input_pdf_name);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%d",&n_x);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%lf %lf %lf", &ymin, &ymax, &yinc);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%s",output_file_name);
			l++;
		}
	}

	/* close parameter file */
	fclose(paramfile);

	/* compute y bin size */
	y_node = floor((ymax - ymin) / yinc + 1.5);

	/* allocate memory */
	x_array = malloc(sizeof(*x_array) * n_x);
	out_array = malloc(sizeof(*out_array) * n_x);
	y_array = malloc(sizeof(*y_array) * y_node);
	hit_matrix = malloc(sizeof(*hit_matrix) * n_x);
	for (i=0; i<n_x; i++) {
		hit_matrix[i] = malloc(sizeof(*hit_matrix) * y_node);
	}

	/* be sure the hit matrix is initialized to 0 */
	for (i=0; i<n_x; i++) {
		for (j=0; j<y_node; j++) {
			hit_matrix[i][j] = 0;
		}
	}

	/* try to open the pdf file */
	if (! (pdf_file = fopen(input_pdf_name,"r"))) {
		fprintf(stderr,"Error file %s not found\n",input_pdf_name);
		exit(1);
	}

	/* read the pdf file input */
	while (fgets(buff,100,pdf_file)) {
		sscanf(buff,"%lf %lf %lf",&x_array[x], &y_array[y], &hit_matrix[x][y]);
		y++;
		if (y == y_node) {
			y=0;
			x++;
		}
	}
		
	/* close the pdf file */
	fclose(pdf_file);

	/* loop over each x point and keep only the y corresponding to the max hit_matrix */
	for (x=0; x<n_x; x++) {
		max_pdf = 0.0;
		for (y=0; y<y_node; y++) {
			if (hit_matrix[x][y] > max_pdf) {
				max_pdf = hit_matrix[x][y];
				max_index = y;
			}
		}
		out_array[x] = y_array[max_index];
	}


	/* open the output file */
	output_file = fopen(output_file_name,"w");

	/* write the output */
	for (i=0; i<n_x; i++) {
		fprintf(output_file,"%lf %lf\n",x_array[i], out_array[i]);
	}

	/* close the output */
	fclose(output_file);

	/* free up memory */
	free(y_array);
//	free(x_array);
//	free(out_array);
//	for (i=0; i<n_x; i++) {
//		free(hit_matrix[i]);
//	}
//	free(hit_matrix);

	/* done message */
	fprintf(stderr,"Done with file: %s\n",output_file_name);

	return 0;
}




