#include <stdio.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
//#include <complex.h>
#include <fftw3.h>
//#include <complex.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include "mysac.h"



#define PI 3.14159265358979311600
#define NPTS_SAC_FILE_MAX 17280000
#define N_MAX_ZEROES 25
#define N_MAX_POLES 25

typedef struct {
        int N;
        double *re;
        double *im;
} SPECTRUM;


typedef struct {
	double real;
	double imag;
} COMPLEX_RP;


typedef struct {
	int n_zeroes;
	int n_poles;
	double scale;
	COMPLEX_RP poles[N_MAX_POLES];
	COMPLEX_RP zeroes[N_MAX_ZEROES];
} POLE_ZERO;


COMPLEX_RP generate_response(POLE_ZERO *pz, double freq);
void decon_response_function(SPECTRUM *data_spect, SPECTRUM *response_spect, SPECTRUM *spect_out);
COMPLEX_RP average_spectral_ratio(SPECTRUM *signal_spectrum,SPECTRUM *untapered_spectrum);
void fftw_forward(double *input, SPECTRUM *spectrum, int npts);
void fftw_backward(SPECTRUM *spectrum, double *array_out);
void cos_taper(double *arr_in, double *arr_out, int npts);
void remove_mean(double *arr_in, double *arr_out, int npts);
void remove_trend(double *arr_in, double *arr_out, int npts);
COMPLEX_RP conjugate(COMPLEX_RP *comp);
COMPLEX_RP complex_multiply(COMPLEX_RP *comp1, COMPLEX_RP *comp2);
COMPLEX_RP complex_divide(COMPLEX_RP *comp1, COMPLEX_RP *comp2);
COMPLEX_RP complex_add(COMPLEX_RP *comp1, COMPLEX_RP *comp2);
COMPLEX_RP complex_subtract(COMPLEX_RP *comp1, COMPLEX_RP *comp2);
COMPLEX_RP complex_exp(COMPLEX_RP *comp);
double complex_amplitude(COMPLEX_RP *comp);
void initialize_spectrum(SPECTRUM *spect, int npts);
void write_sac (char *fname, float *sig, SAC_HD *SHD);
int read_sac (char *fname, float *sig, SAC_HD *SHD, int nmax);
int read_sac_pole_zero_file(char *filename, POLE_ZERO *pz_out);
void zero_pole_zero(POLE_ZERO *pz);
void get_amplitude_spectrum(SPECTRUM *spect_in, double *arr_out);
void time_derivative(double *in, double *out, double delta, int npts);
void bw6plp(double *x, double *arr_out, int np, double dt, double lpfc);
void decimate_signal(double *input_signal, double *arr_out, int npts, double dt_in, double dt_out);
int gcd(int a, int b);
double find_common_sample_rate(double dt_1, double dt_2);
double find_time_shift(int sec_1, int msec_1, int sec_2, int msec_2);
void phase_shift(double t_shift, double dt,int np, SPECTRUM *spect_in);
void signal_to_ground_acceleration(double *in, double *out, int npts, double delta, POLE_ZERO *pz, int acc_flag);
double recursive_mean(double new_value, double old_mean, int n);
double recursive_standard_deviation(double new_value, double current_mean, double prev_value, int n);
void running_mean_psd(double *arr_in, double *arr_out, int npts);
void calculate_psd(SPECTRUM *spect, double *psd_out, double dt, int npts);
void smooth_1d_array(double *arr_in, double *arr_out, int npts, int n_smooth);

