/*
 *  smooth_general_pdf.c
 *  
 *
 *  Created by Rob Porritt on 10/1/10.
 *  Copyright 2010 UC Berkeley. All rights reserved.
 *
 */
#include "station_analysis_tools.h"

int main (int argc, char *argv[]) {
	/* local variables */
	int i, j, l, n_x, y_node, x=0, y=0;
	int max_index=0, nsmooth=5;
	double max_pdf;
	double **hit_matrix;
	double ymin, ymax, yinc;	
	double *x_array, *out_array, *y_array;
	char paramfile_name[100], input_pdf_name[100];
	char output_file_name[100], buff[100];
	FILE *paramfile, *output_file, *pdf_file;
	
	/* help usage */
	if (argc == 2) {
		if (!(strcmp(argv[1],"-h"))) {
			fprintf(stderr,"Usage: %s parameter_file\n", argv[0]);
			fprintf(stderr,"parameter_file format: \n");
			fprintf(stderr,"input_pdf\n");
			fprintf(stderr,"n_x\n");
			fprintf(stderr,"ymin ymax yinc\n");
			fprintf(stderr,"smoothing_radius (nodes)\n");
			fprintf(stderr,"output_file\n");
			exit(1);
		}
	}
	
	/* check usage */
	if (argc != 2) {
		fprintf(stderr,"Usage: %s parameter_file\n", argv[0]);
		fprintf(stderr,"parameter_file format: \n");
		fprintf(stderr,"input_pdf\n");
		fprintf(stderr,"n_x\n");
		fprintf(stderr,"ymin ymax yinc\n");
		fprintf(stderr,"output_file\n");
		fprintf(stderr,"smoothing_radius (nodes)\n");
		exit(1);
	}
	
	/* open parameter file */
	sscanf(argv[1],"%s",paramfile_name);
	if (! (paramfile = fopen(paramfile_name,"r"))) {
		fprintf(stderr,"Error, file %s not found\n",paramfile_name);
		exit(1);
	}
	
	/* read parameter file */
	l=0;
	while (fgets(buff,100,paramfile)) {
		if (l == 0) {
			sscanf(buff,"%s",input_pdf_name);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%d",&n_x);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%lf %lf %lf", &ymin, &ymax, &yinc);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%d", &nsmooth);
			l++;
		} else if (l == 4) {
			sscanf(buff,"%s",output_file_name);
			l++;
		}
	}
	
	/* close parameter file */
	fclose(paramfile);
	
	/* compute y bin size */
	y_node = floor((ymax - ymin) / yinc + 1.5);
	
	/* allocate memory */
	x_array = malloc(sizeof(*x_array) * n_x);
	out_array = malloc(sizeof(*out_array) * n_x);
	y_array = malloc(sizeof(*y_array) * y_node);
	hit_matrix = malloc(sizeof(*hit_matrix) * n_x);
	for (i=0; i<n_x; i++) {
		hit_matrix[i] = malloc(sizeof(*hit_matrix) * y_node);
	}
	
	/* be sure the hit matrix is initialized to 0 */
	for (i=0; i<n_x; i++) {
		for (j=0; j<y_node; j++) {
			hit_matrix[i][j] = 0;
		}
	}
	
	/* try to open the pdf file */
	if (! (pdf_file = fopen(input_pdf_name,"r"))) {
		fprintf(stderr,"Error file %s not found\n",input_pdf_name);
		exit(1);
	}
	
	/* read the pdf file input */
	for (x=0; x<n_x; x++) {
		for (y=0; y<y_node; y++) {
			fscanf(pdf_file,"%lf %lf %lf",&x_array[x],&y_array[y],&hit_matrix[x][y]);
		}
	}
	
	/* close the pdf file */
	fclose(pdf_file);
	
	/* smooth the pdf */
	//int smooth_2d_array(double **dat, double **out, int n_x, int n_y, int n_smooth);
	if ( smooth_2d_array(hit_matrix, hit_matrix, n_x, y_node, nsmooth) == 0) {
		fprintf(stderr,"Error in smooth2Darray - model not smoothed\n");
	}

	/* open up the output file */
	if ( ( output_file = fopen(output_file_name,"w")) == NULL) {
		fprintf(stderr,"Error, could not write to file %s\n",output_file_name);
		exit(1);
	}
	
	/* output the smoothed pdf */
	for (x=0; x<n_x; x++) {
		for (y=0; y<y_node; y++) {
			fprintf(output_file,"%lf %lf %lf\n",x_array[x],y_array[y], hit_matrix[x][y]);
		}
	}
	
	/* close up the output file */
	fclose(output_file);
	
	/* message user its ready */
	fprintf(stdout,"File %s ready.\n",output_file_name);
	
	/* free up memory */
	free(x_array);
	free(y_array);
	for (x=0; x<n_x; x++) {
		free(hit_matrix[x]);
	}
	free(hit_matrix);
	
	return 0;
}
	
	
	
