/* main code to compute a incoherent noise of three time series
 *  
 * follows the definition of Sleeman, 2006 to compute a relative transfer function for instrument self-noise
 * hij = Pik / Pjk
 * nii = Pii - Pij*hij
 * for co-located sensors i, j, k
 * noise spectra n{i,j,k} from power spectra Pii, Pjj, and Pkk
 * conversion to ground motion not necessary, but ok to do
 *
*/

#include "station_analysis_tools.h"


int main (int argc, char *argv[]) {
	/* local variable declaration */
	int i, ii, l, j, acc_flag=1;
	int start_sample, samples_per_second, temp_npts;
	int n_incoherence;
	int common_npts;
	float n_second_sub_record, n_sec_shift;
	FILE *paramfile;
	FILE *output_file;
	char paramfile_name[100];
	char buff[100];
	char output_file_name[100];
	char sac_file_name_1[100], sac_file_name_2[100];
	char pz_file_name_1[100], pz_file_name_2[100];
	float *temp_array_f;
	double low_pass_corner, common_dt, time_shift;
	double *signal_1, *signal_2;
	double *temp_signal, *temp_coherence;
	double *mean_incoherence_1, *sd_incoherence_1;
	double *mean_incoherence_2, *sd_incoherence_2;
	double *incoherence_1, *incoherence_2;
	SAC_HD sac_header_1, sac_header_2;
	fftw_complex *temp_spect_1, *temp_spect_2;
	POLE_ZERO pz_1, pz_2;

	/* temporary arrays now used in signal_to_ground_motion to "simplify" memory management */
	/*
	 * double *signal, double *untapered, SPECTRUM *signal_spectrum, SPECTRUM *response_spectrum,
	 * SPECTRUM *untapered_spectrum, double *fftw_dbl, fftw_complex *fftw_cmplx,
	 * fftw_plan plan_forward, fftw_plan plan_backward
	 */
	double *stga_signal, *stga_untapered, *stga_fftw_dbl;
	fftw_complex *stga_signal_spectrum, *stga_response_spectrum, *stga_untapered_spectrum;
	fftw_complex *stga_fftw_cmplx;
	fftw_plan stga_plan_forward, stga_plan_backward;

	//check for -h usage. Should help prevent misuse.
	if (argc==2) {
		if (!(strcmp(argv[1],"-h"))) {
			fprintf(stderr,"USAGE: %s [parameter_file]\n", argv[0]);
			fprintf(stderr,"parameter file format:\n");
			fprintf(stderr,"sac_file_1\n");
			fprintf(stderr,"polezero_file_1\n");
			fprintf(stderr,"sac_file_2\n");
			fprintf(stderr,"polezero_file_2\n");
			fprintf(stderr,"number_of_seconds_in_each_record number_of_seconds_offset_between_records unit_flag\n");
			fprintf(stderr,"output_file\n\n\n");
			fprintf(stderr,"The full .SAC files will be divided into N records chosen by number_of_seconds_in_each_record and number_of_seconds_offset_between_records.\n");
			fprintf(stderr,"As an example, putting 28800 and 3600 for those parameters for a 24 hour 1 Hz SAC file computes 17 coherence estimates each with length of 8 hours overlapping 1 hour\n");
			fprintf(stderr,"unit_flag can be 0, 1, or 2. This determines the number of derivatives computed - 0 = displacement, 1 = velocity, 2 = acceleration\n");
			fprintf(stderr,"Ideally both files will have a common sampling rate. The programs checks and decimates if necessary.\n");
			exit(1);
		}
	}

	/* open and read the parameter file */
	/* grab command line arguments */
	if(argc!=2) {
		fprintf(stderr,"USAGE: %s [parameter_file]\n", argv[0]);
		fprintf(stderr,"parameter file format:\n");
		fprintf(stderr,"sac_file\n");
		fprintf(stderr,"polezero_file\n");
		fprintf(stderr,"sac_file_2\n");
		fprintf(stderr,"polezero_file_2\n");
		fprintf(stderr,"number_of_seconds_in_each_record number_of_seconds_offset_between_records unit_flag\n");
		fprintf(stderr,"output_file\n\n\n");
		fprintf(stderr,"The full .SAC file will be divided into N records chosen by number_of_seconds_in_each_record and number_of_seconds_offset_between_records.\n");
		fprintf(stderr,"As an example, putting 28800 and 3600 for those parameters for a 24 hour 1 Hz SAC file computes 17 psds estimates each with length of 8 hours overlapping 1 hour\n");
		fprintf(stderr,"unit_flag can be 0, 1, or 2. This determines the number of derivatives computed - 0 = displacement, 1 = velocity, 2 = acceleration\n");
		fprintf(stderr,"Ideally both files will have a common sampling rate. The programs checks and decimates if necessary.\n");
		exit(1);
	}

	// copy the argument to the parameter filename
	strcpy(paramfile_name, argv[1]);

	//open the parameter file
	if (! (paramfile = fopen(paramfile_name,"r"))) {
		fprintf(stderr,"Error: file %s not found\n", paramfile_name);
		exit(1);
	}
        
	// read the parameter file and store its contents
	l=0;
	while (fgets(buff,100,paramfile)) {
		if (l == 0) {
			sscanf(buff,"%s", sac_file_name_1);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%s",pz_file_name_1);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%s",sac_file_name_2);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%s",pz_file_name_2);
			l++;
		} else if (l == 4) {
			sscanf(buff,"%f %f %d",&n_second_sub_record, &n_sec_shift, &acc_flag);
			l++;
  	   	 } else if (l == 5) {
			sscanf(buff,"%s", output_file_name);
			l++;
		} 
	}

	fclose(paramfile);

	//run some checks on the parameters
	//first check is for a reasonable unit to compute
	if (acc_flag != 0 && acc_flag != 1 && acc_flag != 2) {
		fprintf(stderr,"Error, unit_flag must equal 0, 1, or 2. Determines the number of differentiations from the displacement response.\nValue read: %d\n",acc_flag);
		exit(1);
	}
	
	//second check that the sub record and time shift are non-zero
	if (n_second_sub_record <= 0.0 ) {
		fprintf(stderr,"Error, n_second_sub_record must be greater than 0.\n");
		exit(1);
	}
	if (n_sec_shift <= 0.0 ) {
		fprintf(stderr,"Error, n_sec_shift must be greater than 0.\n");
		exit(1);
	}

	//read the first sac file
	temp_array_f = malloc(sizeof(*temp_array_f) * NPTS_SAC_FILE_MAX);
	signal_1 = malloc(sizeof(*signal_1) * NPTS_SAC_FILE_MAX);
	if (!(read_sac ( sac_file_name_1, temp_array_f, &(sac_header_1), NPTS_SAC_FILE_MAX))) {
		fprintf(stderr,"file %s not found\n", sac_file_name_1 );
		exit(1);
	}
	for (i=0; i<sac_header_1.npts; i++) {
		signal_1[i] = (double) temp_array_f[i];
	}
	
	//read the second sac file
	signal_2 = malloc(sizeof(*signal_2) * NPTS_SAC_FILE_MAX);
	if (!(read_sac ( sac_file_name_2, temp_array_f, &(sac_header_2), NPTS_SAC_FILE_MAX))) {
		fprintf(stderr,"file %s not found\n", sac_file_name_2 );
		exit(1);
	}
	for (i=0; i<sac_header_2.npts; i++) {
		signal_2[i] = (double) temp_array_f[i];
	}
	
	//free(temp_array_f);
	
	//check the time steps and decimate if necessary (not appropriate if evalresp library is used - should decimate after instrument deconvolution, but doesn't matter when only using the sensor pole zero)
	//compute a common time step and low pass corner for the decimation
    common_dt = find_common_sample_rate(sac_header_1.delta, sac_header_2.delta);
    low_pass_corner = 0.8 * (1/(2*common_dt));

    //check the sample rates of each and decimate if necessary
    if (sac_header_1.delta != sac_header_2.delta) {
    	bw6plp(signal_1, signal_1, sac_header_1.npts,sac_header_1.delta, low_pass_corner);
        bw6plp(signal_2, signal_2, sac_header_2.npts,sac_header_2.delta, low_pass_corner);
        decimate_signal(signal_1, signal_1, sac_header_1.npts, sac_header_1.delta, common_dt);
        decimate_signal(signal_2, signal_2, sac_header_2.npts, sac_header_2.delta, common_dt);
    }
    
    /* compute the number of points in each and put that into the sac headers */
    if (sac_header_1.npts < sac_header_2.npts) {
    	common_npts = sac_header_1.npts * sac_header_1.delta / common_dt;
    } else {
    	common_npts = sac_header_2.npts * sac_header_2.delta / common_dt;
    }
    	
    sac_header_1.delta = common_dt;
    sac_header_1.npts = common_npts;

    sac_header_2.delta = common_dt;
    sac_header_2.npts = common_npts;


	//read the first pole zero
	if ( read_sac_pole_zero_file(pz_file_name_1, &pz_1) == -1) {
		fprintf(stderr,"Error, file %s not found\n",pz_file_name_1);
		exit(1);
	}

	//read the second pole zero
	if ( read_sac_pole_zero_file(pz_file_name_2, &pz_2) == -1) {
		fprintf(stderr,"Error, file %s not found\n",pz_file_name_2);
		exit(1);
	}

	//invert delta to samples per second
	samples_per_second = floor(1.0/common_dt + 0.5);
	
	//hard wired for 8 hour data segments. (8*3600 = seconds in 8 hours)
	temp_npts = (int) floor(n_second_sub_record * samples_per_second+0.5);
	
	//allocate memory for working arrays used in the next few steps.
	temp_signal = malloc(sizeof(*temp_signal) * temp_npts);
	temp_coherence = malloc(sizeof(*temp_coherence) * temp_npts / 2+1);
	mean_incoherence_1 = malloc(sizeof(*mean_incoherence_1) * temp_npts / 2+1);
	sd_incoherence_1 = malloc(sizeof(*sd_incoherence_1) * temp_npts / 2+1);
	mean_incoherence_2 = malloc(sizeof(*mean_incoherence_2) * temp_npts / 2+1);
	sd_incoherence_2 = malloc(sizeof(*sd_incoherence_2) * temp_npts / 2+1);
	incoherence_1 = malloc(sizeof(*incoherence_1) * temp_npts / 2+1);
	incoherence_2 = malloc(sizeof(*incoherence_2) * temp_npts / 2+1);

	//initialize the data structures
//	initialize_spectrum(&temp_spect_1,temp_npts);  //spectrum 1
//	initialize_spectrum(&temp_spect_2,temp_npts);  //spectrum 2
	temp_spect_1 = fftw_malloc(sizeof(*temp_spect_1) * temp_npts/2+1);
	temp_spect_2 = fftw_malloc(sizeof(*temp_spect_2) * temp_npts/2+1);
	
	//compute the maximum number of coherences to compute
	n_incoherence = ( common_npts - temp_npts ) / (int) (n_sec_shift * (float) samples_per_second) + 1;

	/* allocation for stga:
	 * double *stga_signal, *stga_untapered, *stga_fftw_dbl;
	 * SPECTRUM stga_signal_spectrum, stga_response_spectrum, stga_untapered_spectrum;
	 * fftw_complex *stga_fftw_cmplx;
	 * fftw_plan stga_plan_forward, stga_plan_backward;
	 */
	if ((stga_signal = fftw_malloc(sizeof(*stga_signal) * temp_npts)) == NULL) {
		fprintf(stderr,"Error allocating stga_signal\n");
		exit(1);
	}
	if ((stga_untapered = fftw_malloc(sizeof(*stga_untapered) * temp_npts)) == NULL) {
		fprintf(stderr,"Error allocating stga_untapered\n");
		exit(1);
	}
	if ((stga_fftw_dbl = fftw_malloc(sizeof(*stga_fftw_dbl) * temp_npts)) == NULL) {
		fprintf(stderr,"Error allocating stga_fftw_dbl\n");
		exit(1);
	}
	if ((stga_fftw_cmplx = fftw_malloc(sizeof(*stga_fftw_cmplx) * (temp_npts/2+1))) == NULL) {
		fprintf(stderr,"Error allocating stga_fftw_cmplx\n");
		exit(1);
	}
	
//	initialize_spectrum(&stga_signal_spectrum, temp_npts);
//	initialize_spectrum(&stga_response_spectrum, temp_npts);
//	initialize_spectrum(&stga_untapered_spectrum, temp_npts);
	if ((stga_signal_spectrum = fftw_malloc(sizeof(*stga_signal_spectrum) * (temp_npts/2+1))) == NULL) {
		fprintf(stderr,"Error allocating stga_signal_spectrum\n");
		exit(1);
	}
	if ((stga_response_spectrum = fftw_malloc(sizeof(*stga_response_spectrum) * (temp_npts/2+1))) == NULL) {
		fprintf(stderr,"Error allocating stga_response_spectrum\n");
		exit(1);
	}
	if ((stga_untapered_spectrum = fftw_malloc(sizeof(*stga_untapered_spectrum) * (temp_npts/2+1))) == NULL) {
		fprintf(stderr,"Error allocating stga_untapered_spectrum\n");
		exit(1);
	}
	
	stga_plan_forward = fftw_plan_dft_r2c_1d(temp_npts, stga_fftw_dbl, stga_fftw_cmplx, FFTW_ESTIMATE);
	stga_plan_backward = fftw_plan_dft_c2r_1d(temp_npts, stga_fftw_cmplx, stga_fftw_dbl, FFTW_ESTIMATE);
	
	//find the time shift between the two time series
	time_shift = find_time_shift(sac_header_1.nzsec, sac_header_1.nzmsec, sac_header_2.nzsec, sac_header_2.nzmsec);

	for (j = 0; j < n_incoherence; j++ ) {
		//hard wired to step 1 hour for each data segment
		start_sample = j * n_sec_shift * samples_per_second;
		//make sure we don't go out of bounds
		if (start_sample + temp_npts < common_npts) {
			
			//fill with data from the current segment
			for (ii=0; ii<temp_npts; ii++) { 
				temp_signal[ii] = signal_1[start_sample+ii];
			}
			//convert to acceleration in m/s/s
			signal_to_ground_acceleration(temp_signal, temp_signal, temp_npts, sac_header_1.delta, &(pz_1), acc_flag,
										  stga_signal, stga_untapered, stga_signal_spectrum, stga_response_spectrum,
										  stga_untapered_spectrum, stga_fftw_dbl, stga_fftw_cmplx,
										  stga_plan_forward, stga_plan_backward);
			
			//fft
			for (ii=0; ii<temp_npts; ii++) {
				stga_fftw_dbl[ii] = temp_signal[ii];
			}
			fftw_execute(stga_plan_forward);
			for (ii=0; ii<temp_npts/2+1; ii++) {
				temp_spect_1[ii][0] = stga_fftw_cmplx[ii][0];
				temp_spect_1[ii][1] = stga_fftw_cmplx[ii][1];
			}
			
			//repeat for the second file
			//fill with data from the current segment
			for (ii=0; ii<temp_npts; ii++) { 
				temp_signal[ii] = signal_2[start_sample+ii];
			}
			//convert to acceleration in m/s/s
			signal_to_ground_acceleration(temp_signal, temp_signal, temp_npts, sac_header_2.delta, &(pz_2), acc_flag,
										  stga_signal, stga_untapered, stga_signal_spectrum, stga_response_spectrum,
										  stga_untapered_spectrum, stga_fftw_dbl, stga_fftw_cmplx,
										  stga_plan_forward, stga_plan_backward);
			
			//fft
			for (ii=0; ii<temp_npts; ii++) {
				stga_fftw_dbl[ii] = temp_signal[ii];
			}
			fftw_execute(stga_plan_forward);
			for (ii=0; ii<temp_npts/2+1; ii++) {
				temp_spect_2[ii][0] = stga_fftw_cmplx[ii][0];
				temp_spect_2[ii][1] = stga_fftw_cmplx[ii][1];
			}	
			
			
			//Apply the time shift to the second spectrum
			phase_shift(time_shift, common_dt, temp_npts, temp_spect_2);
			
			//inverse transform to confirm the phase shift worked
			sac_header_1.npts = temp_npts;
			sac_header_2.npts = temp_npts;
			sac_header_2.nzsec = sac_header_1.nzsec;
			sac_header_2.nzmsec = sac_header_1.nzmsec;
			for (ii=0; ii<temp_npts/2+1; ii++) {
				stga_fftw_cmplx[ii][0] = temp_spect_2[i][0];
				stga_fftw_cmplx[ii][1] = temp_spect_2[i][1];
			}
			fftw_execute(stga_plan_backward);
			for (ii=0; ii<temp_npts; ii++) {
				temp_array_f[i] = (float) stga_fftw_dbl[i];
			}
			
			sprintf(buff,"tmpSpect2_%d.SAC",j);
			write_sac(buff,temp_array_f,&sac_header_2);
			
			for (ii=0; ii<temp_npts/2+1; ii++) {
				stga_fftw_cmplx[ii][0] = temp_spect_1[i][0];
				stga_fftw_cmplx[ii][1] = temp_spect_1[i][1];
			}
			fftw_execute(stga_plan_backward);
			for (ii=0; ii<temp_npts; ii++) {
				temp_array_f[i] = (float) stga_fftw_dbl[i];
			}
			
			sprintf(buff,"tmpSpect1_%d.SAC",j);
			write_sac(buff,temp_array_f,&sac_header_1);
	
			//calculate the coherence using the relation Cxy = |Gxy|^2 / (Gxx * Gyy)
			calculate_coherence(temp_spect_1, temp_spect_2, temp_coherence, temp_npts/2+1);

			//compute the incoherence for each spectrum
			calculate_incoherence(temp_coherence, temp_spect_1, incoherence_1, common_dt, temp_npts);
			calculate_incoherence(temp_coherence, temp_spect_2, incoherence_2, common_dt, temp_npts);

			//for each period, compute the mean and standard deviation recursively
			if (j == 0) {
				for (ii=0; ii<temp_npts/2+1; ii++) {
					mean_incoherence_1[ii] = incoherence_1[ii];
					sd_incoherence_1[ii] = 0.0;
					mean_incoherence_2[ii] = incoherence_2[ii];
					sd_incoherence_2[ii] = 0.0;
				}
			} else {	
				for (ii=0; ii<temp_npts/2+1; ii++) {
					mean_incoherence_1[ii] = recursive_mean(incoherence_1[ii], mean_incoherence_1[ii], j+1);
					sd_incoherence_1[ii] = recursive_standard_deviation(incoherence_1[ii], mean_incoherence_1[ii], sd_incoherence_1[ii], j+1);
					mean_incoherence_2[ii] = recursive_mean(incoherence_2[ii], mean_incoherence_2[ii], j+1);
					sd_incoherence_2[ii] = recursive_standard_deviation(incoherence_2[ii], mean_incoherence_2[ii], sd_incoherence_2[ii], j+1);
				}
			}
		}
		
	} // done with each segment

	//simple message to tell the user how many psds were calculated
	fprintf(stdout,"%d incoherences computed and averaged\n",n_incoherence);

	//output
	output_file = fopen(output_file_name,"w");
	for (i=1; i<temp_npts/2+1; i++) {
		fprintf(output_file,"%3.5lf %3.5lf %3.5lf %3.5lf %3.5lf\n" , 1.0 / ( 1.0 / common_dt * i / temp_npts), mean_incoherence_1[i], sd_incoherence_1[i], mean_incoherence_2[i], sd_incoherence_2[i]);
	}
	fclose(output_file);

	
	//free up the memory like a responsible citizen - the OS should do this anyways...
	free(signal_1);
	free(signal_2);
	free(temp_coherence);
	free(mean_incoherence_1);
	free(sd_incoherence_1);
	free(incoherence_1);
	free(mean_incoherence_2);
	free(sd_incoherence_2);
	free(incoherence_2);
	free(temp_signal);
	fftw_free(temp_spect_1);
	fftw_free(temp_spect_2);
	fftw_free(stga_fftw_cmplx);
	fftw_free(stga_fftw_dbl);
	fftw_free(stga_signal);
	fftw_free(stga_untapered);
	fftw_free(stga_signal_spectrum);
	fftw_free(stga_untapered_spectrum);
	fftw_free(stga_response_spectrum);
	fftw_destroy_plan(stga_plan_forward);
	fftw_destroy_plan(stga_plan_backward);
	free(temp_array_f);
	
/*
	free(temp_spect_1.re);
	free(temp_spect_1.im);
	free(temp_spect_2.re);
	free(temp_spect_2.im);
*/
	return 0;
}
