#include "compute_psd.h"


//running mean (general - previous function uses the "k" argument with a function of log10 to be specific to power spectral densities).
//also new is the psd version is directly from david dolenc's code while this is more from rob's smoothing code, just simplified from the 2-d case to the 1-d case.
void smooth_1d_array(double *arr_in, double *arr_out, int npts, int n_smooth)
{
int i, j, start, end, count;
double sum;
if (n_smooth <= 0 || n_smooth >= npts) {
   return;
}

//odd
if (n_smooth % 2 == 1) {
  for (i=0; i<npts; i++) {
    sum = 0.0;
    count = 0;
    if (i-floor(n_smooth/2) < 0) {
      start = 0;
    } else {
      start = i-floor(n_smooth/2);
    }
    if (i+floor(n_smooth/2) > npts-1) {
      end = npts-1;
    } else {
      end = i+floor(n_smooth/2);
    }
    for (j=start; j<=end; j++) {
      sum = sum + arr_in[j];
      count = count + 1;
    }
    arr_out[i] = sum / (double) count;
  }
} else {
  for (i=0; i<npts; i++) {
    sum = 0.0;
    count = 0;
    if (i-floor(n_smooth/2) < 0) {
      start = 0;
    } else {
      start = i-floor(n_smooth/2);
    }
    if (i+floor(n_smooth/2)-1 > npts-1) {
      end = npts-1;
    } else {
      end = i+floor(n_smooth/2)-1;
    }
    for (j=start; j<=end; j++) {
      sum = sum + (arr_in[j] + arr_in[j+1]) / 2.0;
      count = count + 1;
    }
    arr_out[i] = sum / (double) count;
  }
}
return;
}
//END

//compute a mean value recursively
double recursive_mean(double new_value, double old_mean, int n) {
	return (1 / (double) n) * (( (double) n-1) * old_mean + new_value);
}
//END

//compute a standard deviation recursively
double recursive_standard_deviation(double new_value, double current_mean, double prev_value, int n) {
	double temp=0.0, new_sd=0.0;
	if (n > 1) {
		temp = (prev_value * prev_value * ((double) n-2)) + ((double) n/( (double) n-1)) * (current_mean - new_value) * (current_mean - new_value);
		new_sd = sqrt(temp/((double) n-1));
	} else {
		fprintf(stderr,"Recursive_standard_deviation error: n must be greater than 1! n: %d\n",n);
	}
	return new_sd;
}
//END


//////running_mean
void running_mean_psd(double *arr_in, double *arr_out, int npts) {
	int i, k, j;
	double sum=0.0;
	
	for (i=0; i<npts; i++) {
		k=floor(2*(log10(i))*(log10(i))+0.5);
		if (i < 3) {
        	arr_out[i] = arr_in[i];
        } else if (i+k < npts) {
			sum=0.0;
			for (j=-k;j<=k;j++) {
				if (i+j < npts) {
					sum=sum+arr_in[i+j];
				}
			}
			arr_out[i] = sum / (2*k+1);
		} else {
			arr_out[i] = arr_in[i];
        }
    }
    return;
}
///END


//calculate the power spectrum
void calculate_psd(SPECTRUM *spect, double *psd_out, double dt, int npts)
{
	int i;
	for (i=0; i<spect->N; i++) {
		psd_out[i] = 10.0*log10((spect->re[i]*spect->re[i] + spect->im[i]*spect->im[i]) * 2 * dt /  (double) npts);
	}
	running_mean_psd(psd_out, psd_out, spect->N);
	return;
}
//END

//do everything in prep_sac_file....
void signal_to_ground_acceleration(double *in, double *out, int npts, double delta, POLE_ZERO *pz, int acc_flag) {
	int i;
	double *signal, *untapered;
	double freq;
	SPECTRUM signal_spectrum, response_spectrum, untapered_spectrum;
	COMPLEX_RP temp_crp;
	
	/* allocate memory for the incoming sac file. The second array is used later to compute the energy loss due to tapering */
	signal = malloc(sizeof(signal) * npts);
	untapered = malloc(sizeof(untapered) * npts);
	
	/* a few spectrums containing real and imaginary components along with the number of spectrum points */
	initialize_spectrum(&signal_spectrum,npts);
	initialize_spectrum(&response_spectrum,npts);
	initialize_spectrum(&untapered_spectrum,npts);
	
	/* some simple preprocessing (removing a mean and linear trend to remove dc offsets after fft */
	remove_mean(in,signal,npts);
	remove_trend(signal,signal,npts);
	
	/* fill the untapered array. This will be used later again for the energy loss computation */
	for (i=0; i<npts; i++) {
		untapered[i] = signal[i];
	}
	
	/* apply a 10% cosine taper as in peterson 1993 and McNamara and Buland */
	cos_taper(signal,signal,npts);
	
	/* fft */
	fftw_forward(signal,&signal_spectrum,npts);
	fftw_forward(untapered,&untapered_spectrum,npts);
	
	/* calculate to find the average energy loss in the real and imarinary components. This value was assumed constant in the references */
	temp_crp = average_spectral_ratio(&signal_spectrum,&untapered_spectrum);
	
	/* the tapered array is no longer necessary, so be a good citizen and free the memory */
	free(untapered);
	free(untapered_spectrum.re);
	free(untapered_spectrum.im);
	
	/* apply the computed energy difference. The computation gets an absolute value > 1 and we know the energy was lost due to tapering, so we multiply here */
	for (i=0; i<signal_spectrum.N; i++) {
		signal_spectrum.re[i] = signal_spectrum.re[i] * temp_crp.real;
		signal_spectrum.im[i] = signal_spectrum.im[i] * temp_crp.imag;
	}
	
	/* Sequence to compute the frequency domain response of a given pole-zero structure */
	for (i=0; i<response_spectrum.N; i++) {
		freq = (double) (i+1) / (delta * (double) npts);
		temp_crp = generate_response(pz,freq);
		response_spectrum.re[i] = temp_crp.real;
		response_spectrum.im[i] = temp_crp.imag;
	}

	/* doing a deconvolution (complex division) of the response spectrum from the signal spectrum */
	decon_response_function(&signal_spectrum,&response_spectrum,&signal_spectrum);
	
	/* done with the response spectrum */
	free(response_spectrum.re);
	free(response_spectrum.im);
	
	/* return to the time domain for the differentiation */
	fftw_backward(&signal_spectrum,signal);
	
	/* done in the spectral domain */
	free(signal_spectrum.re);
	free(signal_spectrum.im);
	
	/* the response returns displacement. Therefore we differentiate twice to get acceleration */
//	a single derivative seems to return acceleration when comparing with the noise model. Thus give the option for 0, 1, or 2 derivatives
	if (acc_flag == 0) {
		for (i=0; i<npts; i++) {
			out[i] = signal[i];
		}
	} else if (acc_flag == 1) {
		time_derivative(signal, out, delta, npts);
	} else {
		time_derivative(signal, signal, delta, npts);
		time_derivative(signal, out, delta, npts);
	}
	
	/* return the memory */
	free(signal);
	
	return;
}
//END

//phase_shift  
void phase_shift(double t_shift, double dt,int np, SPECTRUM *spect_in) {
	double tpi = 2 * PI, t, w, phi;
	int i, j;
	COMPLEX_RP ca, cx, temp_complex, cb;
	
	t = (double) np * dt;
	
//	for (i=1; i < np / 2+1; i++) {
	for (i=0; i < spect_in->N; i++) {
		temp_complex.real = spect_in->re[i];
		temp_complex.imag = spect_in->im[i];
//		printf("i: %d\n",i);
//		j = i - 1;
		j=i;
		w = tpi * (double) j / t;
		phi = w * t_shift;	
		ca.real = 0.0;
		ca.imag = -1.0*phi;
		cb = complex_exp(&ca);
		cx = complex_multiply(&temp_complex,&cb);
		spect_in->re[i] = cx.real;
		spect_in->im[i] = cx.imag;
	}
	return;
}
//END


///FIND_TIME_SHIFT
double find_time_shift(int sec_1, int msec_1, int sec_2, int msec_2) {
	double time_shift;
//	printf("sec_2: %d msec_2: %d, sec_1: %d, msec_1: %d\n", sec_2, msec_2, sec_1, msec_1);
	time_shift = ( (double) sec_2 + (double) msec_2 / 1000.0) - ( (double) sec_1 + (double) msec_1 / 1000.0);
	return time_shift;	
}
///END

//FIND_COMMON_SAMPLE_RATE
double find_common_sample_rate(double dt_1, double dt_2) {
	double out_rate;
	int sps_1, sps_2;
	
	sps_1 = floor(1/dt_1+.5);
	sps_2 = floor(1/dt_2+.5);
	
	out_rate = 1.0 / gcd(sps_1,sps_2);
	
	return out_rate;
}
//END

//Greatest Common Divisor
int gcd(int a, int b) {
	int out;
	int c;
	
	if (a < b) {
		c=a;
		a=b;
		b=c;
	}
	
	c=1;
	while (c !=0 ) {
		c = a % b;
		if (c!=0) {
			a=b;
			b=c;
		} else {
			out = b;
		}
	}
		
	return out;
}
//END

//decimate_signal
void decimate_signal(double *input_signal, double *arr_out, int npts, double dt_in, double dt_out) {
	int npts_out;
	int i, j;
	
	npts_out = npts * dt_in / dt_out;
	j = 0;
	
	for (i=0; i<npts_out; i++) {
		j = i * dt_out / dt_in;
		if (j<npts) {
			arr_out[i] = input_signal[j];
		} else {
			arr_out[i] = input_signal[npts-1];
		} 
	} 
	
	return;
}
///END	


//6pole low pass butterworth filter - translated from bob urhammer's fortran code
void bw6plp(double *x, double *arr_out, int np, double dt, double lpfc) {
// 6 pole low-pass butterworth filter
/* local variables */
int i, k, nbpp;
double wc, hc, dc, cc11, cc12, cc13;
double cc21, cc22, cc23, cc31, cc32, cc33;
double wim1, xim1, yim1, zim1, wi, xi, yi, zi;
double wim2, xim2, yim2, zim2;

/*initialize some variables. May be a wise decision to make these inputs */
//dt = 1 / 200; /* samples per second ie 200Hz */
//lpfc = 32.0;
nbpp = 3;
wc = tan ( PI * lpfc * dt);

k=0;
hc = sin(PI * (2.0 * (double) k + 1.0) / ( 2.0 * (double) nbpp));
dc = wc * wc + 2.0 * wc * hc + 1.0;
cc11 = wc * wc / dc;
cc12 = (2.0 - 2.0*wc * wc) / dc;
cc13 = ( -1 * wc * wc + 2.0 * wc * hc - 1.0 ) / dc;

k = 1;
hc = sin ( PI * (2.0 * (double) k + 1.0 ) / ( 2.0 * (double) nbpp));
dc = wc * wc + 2.0 * wc * hc + 1.0;
cc21 = wc * wc / dc;
cc22 = (2.0  - 2.0 * wc * wc ) / dc;
cc23 = ( -1 * wc * wc + 2.0 * wc * hc - 1.0) / dc;

k = 2;
hc = sin ( PI * (2.0 * (double) k + 1.0 ) / ( 2.0 * (double) nbpp));
dc = wc * wc + 2.0 * wc * hc + 1.0;
cc31 = wc * wc / dc;
cc32 = (2.0  - 2.0 * wc * wc ) / dc;
cc33 = ( -1 * wc * wc + 2.0 * wc * hc - 1.0) / dc;

/* initialize temporary variables */
wim1 = x[0];
xim1 = x[0];
yim1 = x[0];
zim1 = x[0];
wi = x[0];
xi = x[0];
yi = x[0];
zi = x[0];

/* apply recursive 6PLP filter */
for (i = 0; i < np; i ++) {
  wim2 = wim1;
  wim1 = wi;
  wi = x[i];
  xim2 = xim1;
  xim1 = xi;
  xi = cc11 * ( wi + 2.0 * wim1 + wim2) + cc12 * xim1 + cc13 * xim2;
  yim2 = yim1;
  yim1 = yi;
  yi = cc21 * ( xi + 2.0 * xim1 + xim2) + cc22 * yim1 + cc23 * yim2;
  zim2 = zim1;
  zim1 = zi;
  zi = cc31 * ( yi + 2.0 * yim1 + yim2) + cc32 * zim1 + cc33 * zim2;
  x[i] = zi;
  arr_out[i] = x[i];
}
return;
} // end bw6plp subprogram
///END


//time derivative (as sac 2 point algorithm)
void time_derivative(double *in, double *out, double delta, int npts) {
	int i;
	for (i=0; i<npts-1;i++) {
		out[i] = (in[i+1] - in[i]) / delta;
	}
	//last value is not filled by sac. Here I fill it with the same as the previous value to have a minimal effect on future usage
	out[npts-1] = out[npts-2];
	
	return;
}
//END

//get_amplitude_spectrum
void get_amplitude_spectrum(SPECTRUM *spect_in, double *arr_out) {
	int i;

	for (i=0; i<spect_in->N; i++) {
		arr_out[i] = sqrt(pow(spect_in->re[i],2.0) + pow(spect_in->im[i],2.0));
	}

	return;
}
//END


//to initailize a pole zero structure
void zero_pole_zero(POLE_ZERO *pz) {
	int i;
	pz->scale = 0.0;
	pz->n_zeroes = 0;
	pz->n_poles = 0;
	for (i=0; i<N_MAX_ZEROES;i++) {
		pz->zeroes[i].real = 0.0;
		pz->zeroes[i].imag = 0.0;
	}
	for (i=0; i<N_MAX_POLES;i++) {
		pz->poles[i].real = 0.0;
		pz->poles[i].imag = 0.0;
	}	
	
	return;
}
//END

//generate response
//generates the response for a specific frequency given in hertz
COMPLEX_RP generate_response(POLE_ZERO *pz, double freq) {	
	/* local variables */
	int i;
	double mod_squared;
	COMPLEX_RP omega, denom, num, temp;
	COMPLEX_RP out;
	
	/* initializations */
	freq = freq * 2 * PI;
	omega.real = 0.0;
	omega.imag = freq;
	denom.real = 1.0;
	denom.imag = 1.0;
	num.real = 1.0;
	num.imag = 1.0;
	
	/* compute the complex laplacian */
	for (i=0; i<pz->n_zeroes; i++) {
		temp = complex_subtract(&omega,&pz->zeroes[i]);
		num = complex_multiply(&num,&temp);
	}
	for (i=0; i<pz->n_poles; i++) {
		temp = complex_subtract(&omega,&pz->poles[i]);
		denom = complex_multiply(&denom,&temp);
	}
	
	/* compute the final zeros / poles */
	temp = conjugate(&denom);
	temp = complex_multiply(&temp,&num);
	mod_squared = denom.real * denom.real + denom.imag * denom.imag;
	temp.real = temp.real / mod_squared;
	temp.imag = temp.imag / mod_squared;
	out.real = pz->scale * temp.real;
	out.imag = pz->scale * temp.imag;
	
	return out;	
}
//END



//decon_response_function
void decon_response_function(SPECTRUM *data_spect, SPECTRUM *response_spect, SPECTRUM *spect_out) {
	int i;
	COMPLEX_RP temp_data, temp_response, temp_out;
	
	for (i=0; i<data_spect->N; i++) {
		temp_data.real = data_spect->re[i];
		temp_data.imag = data_spect->im[i];
		temp_response.real = response_spect->re[i];
		temp_response.imag = response_spect->im[i];
		temp_out = complex_divide(&temp_data,&temp_response);
		spect_out->re[i] = temp_out.real;
		spect_out->im[i] = temp_out.imag;
	}
	
	return;

}
///END


//average ratio between two spectrums
COMPLEX_RP average_spectral_ratio(SPECTRUM *signal_spectrum,SPECTRUM *untapered_spectrum) {
	int i;
	double sum_re=0.0, sum_im=0.0;
	double temp;
	COMPLEX_RP ratio;
	
	ratio.real = 0.0;
	ratio.imag = 0.0;

	for (i=0; i<signal_spectrum->N; i++) {
		if (signal_spectrum->re[i] < 0.001 || untapered_spectrum->re[i] < 0.001) {
			temp = 1.0;
		} else {
			temp = fabs(signal_spectrum->re[i] / untapered_spectrum->re[i]);
		}
		sum_re = sum_re + temp;
		if (signal_spectrum->im[i] < 0.001 || untapered_spectrum->im[i] < 0.001) {
			temp = 1.0;
		} else {
			temp = fabs(signal_spectrum->im[i] / untapered_spectrum->im[i]);
		}
		sum_im = sum_im + temp;

	}

	ratio.real = sum_re / (double) signal_spectrum->N;
	ratio.imag = sum_im / (double) signal_spectrum->N;
	
	return ratio;
}
//END


///fftw_forward///
void fftw_forward(double *input, SPECTRUM *spectrum, int npts)
{
  int i;
  double *in;
  int nc;
  fftw_complex *out;
  fftw_plan plan_forward;
  
  in = fftw_malloc ( sizeof (double) * npts);
  for (i=0; i<npts; i++) {
  	in[i] = input[i];
  }

//  free(input);

  nc = spectrum->N;

  out = fftw_malloc ( sizeof ( fftw_complex ) * nc );
  
  plan_forward = fftw_plan_dft_r2c_1d ( npts, in, out, FFTW_ESTIMATE );

  fftw_execute ( plan_forward );
  	
  fftw_destroy_plan ( plan_forward );

  fftw_free ( in );
	
  for (i=0; i<nc; i++) {
	spectrum->re[i] = out[i][0];
	spectrum->im[i] = out[i][1];
  }
	
  fftw_free(out);
  return;
}
////END

///fftw_backward////
void fftw_backward(SPECTRUM *spectrum, double *array_out)
{
  int i;
  double *temp;
  fftw_complex *spect;
  fftw_plan plan_backward;
  int n;

  spect = fftw_malloc( sizeof ( fftw_complex ) * spectrum->N);

  //initialize spect from spectrum
  n = (spectrum->N - 1) * 2;

  for (i=0; i<spectrum->N; i++) {
        spect[i][0] = spectrum->re[i];
        spect[i][1] = spectrum->im[i];
  }

  temp = fftw_malloc ( sizeof (double) * n );
  plan_backward = fftw_plan_dft_c2r_1d(n,spect,temp,FFTW_ESTIMATE);
  
  fftw_execute(plan_backward);
  fftw_destroy_plan(plan_backward);

  fftw_free(spect);
  for ( i = 0; i < n; i++ )
  {
    array_out[i] = temp[i] / (double) n;
  }
  
//  free_spectrum(spectrum); 
  fftw_free ( temp );
	
  return;
}
////END



//calculate a 10% cosine taper
void cos_taper(double *arr_in, double *arr_out, int npts)
{
	int i,M;
	M = floor( ((npts - 1) / 10)/2 + .5);
	
	for (i=0; i<npts; i++) {
	    if (i<M) {
		arr_out[i] = arr_in[i] * (0.5 * (1-cos(i * PI / (M + 1)))) ;
	    } else if (i<npts - M - 2) {
		arr_out[i] = arr_in[i];
	    } else if (i<npts) {
		arr_out[i] = arr_in[i] * ( 0.5 * (1-cos((npts - i - 1) * PI / (M + 1))));
	    }
	}
	return;
	//http://www.cbi.dongnocchi.it/glossary/DataWindow.htm
}
/////END

//remove_mean
void remove_mean(double *arr_in, double *arr_out, int npts)
{
	int i;
	double ave=0.0, sum=0.0;
	
	for (i=0; i<npts; i++) {
		sum = sum + arr_in[i];
	}
	ave = sum / (double) npts;
	for (i=0; i<npts;i++) {
		arr_out[i] = arr_in[i] - ave;
	} 
	return;
}
///END

void remove_trend(double *arr_in, double *arr_out, int npts)
{
	int i;
	double ata11, ata12, ata21, ata22, atb1, atb2;
	double fn, di, d, a, b;
	double ata11i, ata12i, ata21i, ata22i;
//	double mean, sum=0.0;
	
/* removes trend with a linear least squares approach */
ata11 = 0.0;
ata12 = 0.0;
ata21 = 0.0;
ata22 = 0.0;
atb1 = 0.0;
atb2 = 0.0;
fn = (double) npts * 1.0;

for (i=0; i<npts; i++) {
	di = (double) i;
	ata11 = ata11 + 1.0;
	ata12 = ata12 + di;
	ata21 = ata21 + di;
	ata22 = ata22 + di * di;
	atb1 = atb1 + arr_in[i];
	atb2 = atb2 + arr_in[i] * di;
//	sum = sum + arr_in[i];
}
//mean = sum / ((double) npts);

d = ata11 * ata22 - ata12 * ata21;
ata11i = ata22 / d;
ata12i = -1 * ata12 / d;
ata21i = -1 * ata21 / d;
ata22i = ata11 / d;
a = ata11i * atb1 + ata12i * atb2;
b = ata21i * atb1 + ata22i * atb2;

//printf("a: %lf, b: %lf\nmean: %lf\n",a,b, mean);

for (i=0; i<npts; i++) {
	arr_out[i] = arr_in[i] - (a + b * (double)i);
}
return;
} /* end linear detrend */
////END



//complex math
COMPLEX_RP conjugate(COMPLEX_RP *comp) {
	COMPLEX_RP comp_out;
	comp_out.real = comp->real;
	comp_out.imag = -1.0 * comp->imag;
	return comp_out;
}

COMPLEX_RP complex_multiply(COMPLEX_RP *comp1, COMPLEX_RP *comp2) {
	COMPLEX_RP comp_out;
	comp_out.real = comp1->real * comp2->real - comp1->imag * comp2->imag;
	comp_out.imag = comp1->real * comp2->imag + comp2->real * comp1->imag;
	return comp_out;	
}

COMPLEX_RP complex_divide(COMPLEX_RP *comp1, COMPLEX_RP *comp2) {
	COMPLEX_RP comp_out;
	double denom;
	denom = comp2->real * comp2->real + comp2->imag * comp2->imag;// + 1.2e-32;
	comp_out.real = (comp1->real*comp2->real + comp1->imag*comp2->imag)/(denom);
	comp_out.imag = (comp1->imag*comp2->real - comp1->real*comp2->imag)/(denom);
	return comp_out; 
}


COMPLEX_RP complex_add(COMPLEX_RP *comp1, COMPLEX_RP *comp2) {
	COMPLEX_RP comp_out;
	comp_out.real = comp1->real + comp2->real;
	comp_out.imag = comp1->imag + comp2->imag;
	return comp_out;
}

COMPLEX_RP complex_subtract(COMPLEX_RP *comp1, COMPLEX_RP *comp2) {
	COMPLEX_RP comp_out;
	comp_out.real = comp1->real - comp2->real;
	comp_out.imag = comp1->imag - comp2->imag;
	return comp_out;
}

COMPLEX_RP complex_exp(COMPLEX_RP *comp) {
	COMPLEX_RP comp_out;
	comp_out.real = exp(comp->real) * cos(comp->imag);
	comp_out.imag = exp(comp->real) * sin(comp->imag);	
	return comp_out;
}

double complex_amplitude(COMPLEX_RP *comp) {
	return sqrt( comp->real * comp->real + comp->imag * comp->imag );
}


//initialize_spectrum
void initialize_spectrum(SPECTRUM *spect, int npts) {
	spect->N = npts / 2 + 1;
	spect->re = malloc(sizeof(spect->re) * spect->N);
	spect->im = malloc(sizeof(spect->im) * spect->N);
	return;
}


//write a sac file
void write_sac (char *fname, float *sig, SAC_HD *SHD)
{
 FILE *fsac;
 int i;

/*..........................................................................*/
        
        fsac = fopen(fname, "wb");

        if ( !SHD ) SHD = &SAC_HEADER;


        SHD->iftype = (long)ITIME;
        SHD->leven = (long)TRUE;

        SHD->lovrok = (long)TRUE;
        SHD->internal4 = 6L;

//printf("pre: %f %f\n",SHD->depmin, SHD->depmax);

  /*+++++++++++++++++++++++++++++++++++++++++*/
     SHD->depmin = sig[0];
     SHD->depmax = sig[0];
//printf("init: %f %f %f\n",SHD->depmin, SHD->depmax, sig[0]);
   for ( i = 0; i < SHD->npts ; i++ )
   {
    if ( SHD->depmin > sig[i] ) SHD->depmin = sig[i];
    if ( SHD->depmax < sig[i] ) SHD->depmax = sig[i];
   }

//	printf("post: %f %f\n",SHD->depmin, SHD->depmax);
         fwrite(SHD,sizeof(SAC_HD),1,fsac);

         fwrite(sig,sizeof(float),(int)(SHD->npts),fsac);


         fclose (fsac);
}
////END




//read a sac file
int read_sac (char *fname, float *sig, SAC_HD *SHD, int nmax) {
		FILE *fsac;
        fsac = fopen(fname, "rb");
        if ( !fsac )
        {
         //fclose (fsac);
         return 0;
        }

        if ( !SHD ) SHD = &SAC_HEADER;

         fread(SHD,sizeof(SAC_HD),1,fsac);

         if ( SHD->npts > nmax )
         {
          fprintf(stderr,
           "ATTENTION !!! in the file %s the number of points is limited to %d\n",fname,nmax);

          SHD->npts = nmax;
         }

         fread(sig,sizeof(float),(int)(SHD->npts),fsac);

        fclose (fsac);

   /*-------------  calculate the initial time  ----------------*/
   {
        int eh, em ,i;
        float fes;
        char koo[9];

        for ( i = 0; i < 8; i++ ) koo[i] = SHD->ko[i];
        koo[8] = '\0';

        SHD->o = SHD->b + SHD->nzhour*3600. + SHD->nzmin*60 +
         SHD->nzsec + SHD->nzmsec*.001;

        sscanf(koo,"%d%*[^0123456789]%d%*[^.0123456789]%g",&eh,&em,&fes);

        SHD->o  -= (eh*3600. + em*60. + fes);
   /*-------------------------------------------*/}
        return 1;
}
//END




//read a pole zero file
int read_sac_pole_zero_file(char *filename, POLE_ZERO *pz_out) {
	char buff[100], temp[100];
	int i=-1;
	int ival;
	int n_zeros, n_poles;
	int zero_scan_flag = 0, pole_scan_flag = 0;
	double fval1, fval2;
	double constant;
	FILE *pz_file;	

	//initialize the pole zero file
	zero_pole_zero(pz_out);

	//try to open the file. return -1 as failure value
	if (! (pz_file = fopen(filename, "r"))) {
		return -1;
	}

	while (fgets(buff,100,pz_file)) {
		sscanf(buff,"%s %d\n", temp, &ival);
		if (!strcmp(temp,"ZEROS")) {
			zero_scan_flag = 1;
			pole_scan_flag = 0;
			i=-1;
			n_zeros = ival;
			pz_out->n_zeroes = n_zeros;
		} else if (!strcmp(temp,"ZEROES")) {
			zero_scan_flag = 1;
			pole_scan_flag = 0;
			i=-1;
			n_zeros = ival;
			pz_out->n_zeroes = n_zeros;
		} else if (! strcmp(temp,"POLES")) {
			zero_scan_flag = 0;
			pole_scan_flag = 1;
			i=-1;
			n_poles = ival;
			pz_out->n_poles = n_poles;
		} else if (! strcmp(temp,"CONSTANT")) {
			zero_scan_flag = 0;
			pole_scan_flag = 0;
			i=-1;
			sscanf(buff,"%s %lf",temp,&constant);
			pz_out->scale = constant;
		} else if (zero_scan_flag == 1) {
			i++;
			sscanf(buff,"%lf %lf", &fval1, &fval2);
			pz_out->zeroes[i].real = fval1;
			pz_out->zeroes[i].imag = fval2;
		} else if (pole_scan_flag == 1) {
			i++;
			sscanf(buff,"%lf %lf", &fval1, &fval2);
			pz_out->poles[i].real = fval1;
			pz_out->poles[i].imag = fval2;
		}
	}
	
	fclose(pz_file);
	
	return 1;
}
//END


