/*
 *  interpolate_psd.c
 *  
 *
 *  Created by Rob Porritt on 10/1/10.
 *  Copyright 2010 UC Berkeley. All rights reserved.
 *
 *  Simple code to interpolate a power spectrum
 *  Basically reads a psd file from compute_psd
 *  and then takes the output period array and
 *  converts it into some regular array which is
 *  user specified
 *
 */


//subroutine for interpolate.
//freqs is an array of frequencies for the input array tmp with nfreqs
//outputs to array out with nfreq number of frequencies
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include "spline.h"

void interpolate_psd(double *freqs, int nfreqs, double *input, double *out, double minPer, double maxPer, double incPer);
int compute_nodes(double max, double min, double inc);
void spline_interpolate_psd(double *freqs, int nfreqs, double *input, double *freqsOut, double *out, int *nFreqsOut, double minPer, double maxPer, double incPer, SPLINECOMMON *sp);
int compute_log10_nodes(int max, int min);
void log10_linear_interpolate(double *freqs, int nfreqs, double *input, double *freqsOut, double *out, int *nFreqsOut, double minPer, double maxPer, double incPer);

//note it was copied from the spline code, so I didn't alter variable names
void log10_linear_interpolate(double *freqs, int nfreqs, double *input, double *freqsOut, double *out, int *nFreqsOut, double minPer, double maxPer, double incPer) {
	int i,j, k, l, nspline;
	int khi, klo;
	double slope, df;
	int tmpMinMax, splineMin, splineMax;
	double *splineFreqs, splineFreq;
	double minFreqIn, maxFreqIn, tmpFreqIn;
	double minval, maxval, tmpval;

	//initialize output
	for (i=0; i<nfreqs; i++) {
		out[i] = 0.0;
		freqsOut[i] = 0.0;
	}

    //initialize the spline parameters
    splineMin = (int) floor(log10(freqs[0])+0.5);
    splineMax = (int) ceil(log10(freqs[nfreqs-1])-0.5);
    if (splineMin > splineMax) {
		tmpMinMax = splineMin;
		splineMin = splineMax;
		splineMax = tmpMinMax;
	}
	minFreqIn = freqs[0];
	maxFreqIn = freqs[nfreqs-1];
	if (maxFreqIn < minFreqIn) {
		tmpFreqIn = minFreqIn;
		minFreqIn = maxFreqIn;
		maxFreqIn = tmpFreqIn;
	}
	minval = input[0];
	maxval = input[nfreqs-1];
	if (minval > maxval) {
		tmpval = minval;
		minval = maxval;
		maxval = tmpval;
	}

	nspline = compute_log10_nodes(splineMax, splineMin);
	*nFreqsOut = nspline;

	//loop out
	k=0;
	klo = 0;
	khi = 1;
	slope = (input[khi] - input[klo]) / (freqs[khi] - freqs[klo]);
	for (i=splineMin; i<=splineMax; i++) {
		for (j=1; j<10; j++) {
			for (l=0; l<10; l++) {
				freqsOut[k] = (double) j*pow(10.0,(double) i) + (double) l * pow(10.0,(double)i-1);
					if (freqsOut[k] >= minFreqIn && freqsOut[k] <= maxFreqIn) {
						while (freqsOut[k] > freqs[khi]) {
							khi++;
							if (khi == nfreqs-1) {
								break;
							}
						}
						klo = khi-1;
						while (freqsOut[k] < freqs[klo]) {
							klo--;
							if (klo == 0) {
								break;
							}
						}
						df = freqsOut[k] - freqs[klo];
						slope = (input[khi] - input[klo]) / (freqs[khi] - freqs[klo]);
						out[k] = input[klo] + df * slope;
					} else if (freqsOut[k] < minFreqIn) {
						out[k] = minval;
					} else {
						out[k] = maxval;
					}
				k++;
			}
		}
	}
	return;
}



int compute_log10_nodes(int max, int min) {
	int i,j,l,k;
	k=0;
	for (i=min; i<max; i++) {
		for (j=1; j<10; j++) {
			for (l=0; l<10; l++) {
				k++;
			}
		}
	}
	return k;
}


//works similar to interpolate_psd, but instead calls spline subroutines
void spline_interpolate_psd(double *freqs, int nfreqs, double *input, double *freqsOut, double *out, int *nFreqsOut, double minPer, double maxPer, double incPer, SPLINECOMMON *sp) {
	int i,j, k, l, nspline, ip=2, ierr;
	int tmpMinMax, splineMin, splineMax;
	double Vpred, s, ss;
	double *splineFreqs, splineFreq;
	double minFreqIn, maxFreqIn, tmpFreqIn;
	double minval, maxval, tmpval;

	//initialize output
	for (i=0; i<nfreqs; i++) {
		out[i] = 0.0;
		freqsOut[i] = 0.0;
	}

	//initialize the spline parameters
	splineMin = (int) floor(log10(freqs[0]));
	splineMax = (int) ceil(log10(freqs[nfreqs-1]));
	if (splineMin > splineMax) {
		tmpMinMax = splineMin;
		splineMin = splineMax;
		splineMax = tmpMinMax;
	}
	minFreqIn = freqs[0];
	maxFreqIn = freqs[nfreqs-1];	
	if (maxFreqIn < minFreqIn) {
		tmpFreqIn = minFreqIn;
		minFreqIn = maxFreqIn;
		maxFreqIn = tmpFreqIn;
	}
	minval = input[0];
	maxval = input[nfreqs-1];
	if (minval > maxval) {
		tmpval = minval;
		minval = maxval;
		maxval = tmpval;
	}

	nspline = compute_log10_nodes(splineMax, splineMin);
	*nFreqsOut = nspline;

//	allocate_spline(&sp,nspline,nfreqs);
	//initial spline point
	ierr = mspline(ip+2, nfreqs, freqs, input, 0, 0.0, 0, 0.0, sp);
	//loop out
	k=0;
        for (i=splineMin; i<=splineMax; i++) {
		for (j=1; j<10; j++) {
			for (l=0; l<10; l++) {
				freqsOut[k] = (double) j*pow(10.0,(double) i) + (double) l * pow(10.0,(double)i-1);
				if (freqsOut[k] >= minFreqIn && freqsOut[k] <= maxFreqIn) {
			                ierr = msplder(ip+2,freqsOut[k],&Vpred,&s,&ss, sp);
			                out[k] = Vpred;
				} else if (freqsOut[k] < minFreqIn) {
					out[k] = minval;
				} else {
					out[k] = maxval;
				}
				k++;
			}
	        }
	}
//	free_spline(&sp,nspline,nfreqs);
	return;
}

//simple node computer
int compute_nodes(double max, double min, double inc) {
	return floor((max-min)/inc+1.5);
}
//END
	
//primary interpolation routine (consider adjusting to use an output frequency array which is given as an argument)
void interpolate_psd(double *freqs, int nfreqs, double *input, double *out, double minPer, double maxPer, double incPer) {

	int i, klo, khi, nfreq;
	double slope, F, incr, df;
	
	nfreq = compute_nodes(maxPer,minPer,incPer);
	F = minPer;
	incr = incPer;
	klo = 0;
	khi = 1;
	slope = (input[khi] - input[klo]) / (freqs[khi] - freqs[klo]);
	for (i=0; i<nfreq; i++) {
		while (F > freqs[khi]) {
			khi++;
			if (khi == nfreqs-1) {
				break;
			}
		}
		klo = khi-1;
		while (F < freqs[klo]) {
			klo--;
			if (klo == 0) {
				break;
			}
		}
		df = F-freqs[klo];
		slope = (input[khi] - input[klo]) / (freqs[khi] - freqs[klo]);
		out[i] = input[klo] + df * slope;
		//if (out[i] < -170.0) {
		//	printf("F: %lf out[%d]: %lf input[%d]: %lf df: %lf slope: %lf input[%d]: %lf\n",F, i, out[i], klo, input[klo], df, slope, khi, input[khi]);
		//}
		F = F+incr;
	}
	
	return;
}
//END


int main (int argc, char *argv[]) {
	/* local variables */
	int x, l, xNode, xNodeOut, spmin, spmax;
	double minPeriodOut, maxPeriodOut, periodIncrement;
	double *frequencyArray, *powerSpectralDensityRead, *powerSpectralDensityOut;
	double *standardDeviationRead, *standardDeviationOut;
	char paramFileName[100], inputFileName[100], outputFileName[100];
	char buff[100];
	FILE *paramFile, *inputFile, *outputFile;
	SPLINECOMMON spc;
	
	/* check usage */
	if ( argc != 2) {
		fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
		fprintf(stderr,"Parameter file format:\n");
		fprintf(stderr,"input_psd_file_name\n");
		fprintf(stderr,"number_of_nodes_in_x\n");
		fprintf(stderr,"min_period_out max_period_out period_increment\n");
		fprintf(stderr,"output_psd_file_name\n");
		exit(1);
	}
	
	/* -h help */
	if ( !(strcmp(argv[1],"-h"))) {
		fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
		fprintf(stderr,"Parameter file format:\n");
		fprintf(stderr,"input_psd_file_name\n");
		fprintf(stderr,"number_of_nodes_in_x\n");
		fprintf(stderr,"min_period_out max_period_out period_increment\n");
		fprintf(stderr,"output_psd_file_name\n");
		exit(1);
	}
	
	/* try to open the input file */
	sscanf(argv[1],"%s",paramFileName);
	if ( ( paramFile = fopen(paramFileName,"r")) == NULL) {
		fprintf(stderr,"Error, file %s not found\n",paramFileName);
		exit(1);
	}
	
	/* read the parameter file */
	l=0;
	while (fgets(buff,100,paramFile)) {
		if (l == 0) {
			sscanf(buff,"%s",inputFileName);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%d",&xNode);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%lf %lf",&minPeriodOut, &maxPeriodOut);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%s",outputFileName);
			l++;
		}
	}

	/* close it up */
	fclose(paramFile);
	
	/* compute the output nodes */
	/* call the interpolation routine */
	int nfreqspline, minspline, maxspline, tmpspline;
	minspline = (int) floor(log10(minPeriodOut)+0.5);
	maxspline = (int) floor(log10(maxPeriodOut)+0.5);
	if (maxspline < minspline) {
		tmpspline = minspline;
		minspline = maxspline;
		maxspline = tmpspline;
	}
	nfreqspline = compute_log10_nodes(maxspline, minspline);
	double *splineFreqs;
	splineFreqs = malloc(sizeof(*splineFreqs) * nfreqspline);
	xNodeOut = nfreqspline;
	
	/* allocate memory */
	if ( ( frequencyArray = malloc(sizeof(*frequencyArray) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating frequency array\n");
		exit(1);
	}
	
	if ( ( powerSpectralDensityRead = malloc(sizeof(*powerSpectralDensityRead) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating psd array\n");
		exit(1);
	}
	
	if ( ( powerSpectralDensityOut = malloc(sizeof(*powerSpectralDensityOut) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating psd output array\n");
		exit(1);
	}
	
	if ( ( standardDeviationRead = malloc(sizeof(*standardDeviationRead) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating sd array\n");
		exit(1);
	}
	
	if ( ( standardDeviationOut = malloc(sizeof(*standardDeviationOut) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating sd output array\n");
		exit(1);
	}
	
	/* try to open the input file */
	if ( ( inputFile = fopen(inputFileName,"r")) == NULL) {
		fprintf(stderr,"Error, could not read inputFileName: %s\n",inputFileName);
		exit(1);
	}
		
	/* read the file */
	for (x=xNode-1; x>=0; x--) {
		fscanf(inputFile,"%lf %lf %lf",&frequencyArray[x],&powerSpectralDensityRead[x],&standardDeviationRead[x]);
	}
	
	/* close the file */
	fclose(inputFile);
		
	/* call the interpolation routine */
	//interpolate_psd(frequencyArray, xNode, powerSpectralDensityRead,powerSpectralDensityOut, minPeriodOut, maxPeriodOut,periodIncrement);
	//interpolate_psd(frequencyArray, xNode, standardDeviationRead,standardDeviationOut, minPeriodOut, maxPeriodOut, periodIncrement);
	
	/* trial with spline */
	
	
	log10_linear_interpolate(frequencyArray, xNode, powerSpectralDensityRead, splineFreqs, powerSpectralDensityOut, &nfreqspline, minPeriodOut, maxPeriodOut,periodIncrement);
	log10_linear_interpolate(frequencyArray, xNode, standardDeviationRead, splineFreqs, standardDeviationOut,&nfreqspline, minPeriodOut, maxPeriodOut, periodIncrement);
	
	/* open the output file */
	if ( ( outputFile = fopen(outputFileName,"w")) == NULL) {
		fprintf(stderr,"Error, could not create file %s\n",outputFileName);
		exit(1);
	}
	
	/* print */
	for (x=nfreqspline-1; x>=0; x--) {
		fprintf(outputFile,"%3.8g %3.8g %3.8g\n",splineFreqs[x],powerSpectralDensityOut[x],standardDeviationOut[x]);
	}
	
	/* close up */
	fclose(outputFile);

	allocate_spline(&spc, nfreqspline, xNode);
	
	spline_interpolate_psd(frequencyArray, xNode, powerSpectralDensityRead, splineFreqs, powerSpectralDensityOut, &nfreqspline, minPeriodOut, maxPeriodOut,periodIncrement, &spc);
	spline_interpolate_psd(frequencyArray, xNode, standardDeviationRead, splineFreqs, standardDeviationOut,&nfreqspline, minPeriodOut, maxPeriodOut, periodIncrement, &spc);

	free_spline(&spc, nfreqspline, nfreqspline);
					
	/* open the output file */
    if ( ( outputFile = fopen("Spline.dat","w")) == NULL) {
		fprintf(stderr,"Error, could not create file Spline.dat\n");
		exit(1);
	}

	/* print */
	for (x=nfreqspline-1; x>=0; x--) {
		fprintf(outputFile,"%3.8g %3.8g %3.8g\n",splineFreqs[x],powerSpectralDensityOut[x],standardDeviationOut[x]);
	}
    /* close up */
	fclose(outputFile);

	/* because we write to a hard wired file the spline part, let the user know */
	printf("log10 based linear interpolated file: %s\nSpline interpolated log10 based file: Spline.dat\n",outputFileName);
	
	/* why not free up memory */
	free(frequencyArray);
	free(splineFreqs);
	free(powerSpectralDensityOut);
	free(standardDeviationOut);
	free(powerSpectralDensityRead);
	free(standardDeviationRead);
	
	return 0;
}








